float g_afk_time = 60;

void print(string text) { g_Game.AlertMessage( at_console, text); }
void println(string text) { print(text + "\n"); }

class PlayerState
{
	string lastHole = "t_long1_ball";
	array<EHandle> swingers; // anyone who swung at the ball this hole
	int swings = 0;
	
	int lastButtons = -1;
	Vector lastAngles;
	float lastNotAfk = 0;
	
	EHandle h_arrow;
	
	bool showStats = false;
	
	void addSwinger(CBasePlayer@ plr) {
		for (uint i = 0; i < swingers.size(); i++) {
			if (!swingers[i].IsValid()) {
				continue;
			}
			
			if (swingers[i].GetEntity().entindex() == plr.entindex()) {
				return;
			}
		}
		
		swingers.insertLast(EHandle(plr));
	}
	
	void afkPoll(CBasePlayer@ plr) {
		if (plr.pev.v_angle != lastAngles || plr.pev.button != lastButtons) {
			lastNotAfk = g_Engine.time;
		}
		lastButtons = plr.pev.button;
		lastAngles = plr.pev.v_angle;
	}
	
	array<CBasePlayer@> getSwingPlayers() {
		array<CBasePlayer@> swingPlrs;
		
		for (uint i = 0; i < swingers.size(); i++) {
			if (!swingers[i].IsValid()) {
				continue;
			}
			
			CBasePlayer@ plr = cast<CBasePlayer@>(swingers[i].GetEntity());
			if (plr !is null && plr.IsConnected())
				swingPlrs.insertLast(plr);
		}
		
		return swingPlrs;
	}
	
	void reset(CBasePlayer@ plr) {
		swings = 0;
		swingers.resize(0);
		lastHole = "t_long1_ball";
		plr.pev.targetname = "player_ball";
		
		g_EntityFuncs.Remove(h_arrow);
		
		dictionary keys;
		keys["origin"] = plr.pev.origin.ToString();
		keys["model"] = "models/minigolf/radarrow.mdl";
		keys["targetname"] = "arrow_hat";
		keys["rendermode"] = "0";
		keys["spawnflags"] = "1";
		CBaseEntity@ arrow = g_EntityFuncs.CreateEntity("env_sprite", keys, true);	

		arrow.pev.movetype = MOVETYPE_FOLLOW;
		@arrow.pev.aiment = plr.edict();
		arrow.pev.framerate = 1;
		
		h_arrow = arrow;
		
		plr.RemoveAllItems(false);
		plr.SetItemPickupTimes(0);
		
		for (int i = 0; i < 64; i++)
			plr.m_rgAmmo(i, 1000);
		
		plr.GiveNamedItem("weapon_saw");
		plr.GiveNamedItem("weapon_medkit");
		//arrow.pev.colormap = plr.pev.colormap;
	}
	
	int getHoleNum() {
		string hole = lastHole.SubString(6);
		hole = hole.SubString(0, hole.Find("_"));
		return atoi(hole)-1;
	}
	
	void finishHole(CBasePlayer@ ball, string nextHole) {
		string ballName = ball.pev.netname;
		
		int holeNum = getHoleNum()+1;
		if (holeNum == 18 || holeNum == 0) {
			lastHole = nextHole;
			swings = 0;
			swingers.resize(0);
			return; // going into the gopher hole
		}
		
		array<CBasePlayer@> swingPlrs = getSwingPlayers();
		
		if (swingPlrs.size() > 0) {
			if (swings <= 1) {
				swingPlrs[0].pev.frags += 20;
				ball.pev.frags += 20;
				
				g_SoundSystem.PlaySound(swingPlrs[0].edict(), CHAN_ITEM, "minigolf/fanfare.wav", 1.0f, 1.0f, 0, 100);
				g_SoundSystem.PlaySound(ball.edict(), CHAN_ITEM, "minigolf/fanfare.wav", 1.0f, 1.0f, 0, 100);
				
				dictionary keys;
				keys["origin"] = swingPlrs[0].pev.origin.ToString();
				keys["spawnflags"] = "3";
				
				CBaseEntity@ ent = g_EntityFuncs.CreateEntity("env_funnel", keys, true);
				ent.Use(null, null, USE_ON);
				ent.pev.origin = ball.pev.origin;
				ent.Use(null, null, USE_ON);
				g_EntityFuncs.Remove(ent);
				
				g_PlayerFuncs.SayTextAll(ball, "Hole in one! (" + swingPlrs[0].pev.netname + " + " + 
					ball.pev.netname + " at hole " + holeNum + ")\n");
			}
			else {
				int addScore = 0;
				if (swings == 2) {
					addScore = 10;
				} else if (swings == 3) {
					addScore = 5;
				} else if (swings == 4) {
					addScore = 2;
				} else if (swings == 5) {
					addScore = 1;
				}
				
				string names = "";
				for (int i = 0; i < int(swingPlrs.size()); i++) {
					swingPlrs[i].pev.frags += addScore;
					names += swingPlrs[i].pev.netname;
					
					if (i < int(swingPlrs.size())-2) {
						names += ", ";
					}
					else if (i < int(swingPlrs.size())-1) {
						names += " and ";
					}
				}
				ball.pev.frags += addScore;
				g_PlayerFuncs.ClientPrintAll(HUD_PRINTNOTIFY, ballName + " made hole " + holeNum + " after " + swings + " swings by " + names +"\n");
				
				if (swings <= 3) {
					for (uint i = 0; i < swingPlrs.size(); i++) {
						swingPlrs[i].pev.frags += addScore;
						g_SoundSystem.PlaySound(swingPlrs[i].edict(), CHAN_ITEM, "minigolf/ding.wav", 1.0f, 1.0f, 0, 100);
					}
					g_SoundSystem.PlaySound(ball.edict(), CHAN_ITEM, "minigolf/ding.wav", 1.0f, 1.0f, 0, 100);
				}
			}
		}
		
		swings = 0;
		lastHole = nextHole;
		swingers.resize(0);
	}
}

dictionary player_states;

PlayerState@ getPlayerState(CBasePlayer@ plr)
{
	string steamId = g_EngineFuncs.GetPlayerAuthId( plr.edict() );
	if (steamId == 'STEAM_ID_LAN') {
		steamId = plr.pev.netname;
	}
	
	if ( !player_states.exists(steamId) )
	{
		PlayerState state;
		player_states[steamId] = state;
	}
	
	PlayerState@ state = cast<PlayerState@>( player_states[steamId] );
	
	return state;
}

void MapInit() {
	g_Hooks.RegisterHook( Hooks::Player::ClientSay, @ClientSay );
	g_Hooks.RegisterHook( Hooks::Player::PlayerTakeDamage, @PlayerTakeDamage );
	g_Hooks.RegisterHook( Hooks::Player::PlayerPreThink, @PlayerPreThink );

	//g_Scheduler.SetInterval("ball_physics_loop", 0, -1);
	g_Scheduler.SetInterval("status_loop", 1, -1);
}

void status_loop() {
	array<int> golfer_count(20);
	array<int> ball_count(20);

	for ( int i = 1; i <= g_Engine.maxClients; i++ )
	{
		CBasePlayer@ plr = g_PlayerFuncs.FindPlayerByIndex(i);
		if (plr is null or !plr.IsConnected())
			continue;
		
		PlayerState@ state = getPlayerState(plr);
		
		state.afkPoll(plr);
		int timeLeft = int(g_afk_time - (g_Engine.time - state.lastNotAfk) + 0.5f);
		bool isAfk = timeLeft <= 0;	
		
		int holeNum = state.getHoleNum();
		if (holeNum >= 0 && holeNum < 20) {
			if (plr.pev.targetname == "golfer" || plr.pev.targetname == "pong_golfer") {
				golfer_count[holeNum]++;
			}
			if (plr.pev.targetname == "player_ball" || plr.pev.targetname == "pong_ball") {
				ball_count[holeNum]++;
			}
		}
		
		if (plr.pev.targetname == "golfer" || plr.pev.targetname == "pong_golfer") {
			if (isAfk) {
				plr.TakeDamage(null, null, 99999, 0);
				plr.pev.targetname = "player_ball";
				g_PlayerFuncs.SayTextAll(plr, "" + plr.pev.netname + " became a ball after being AFK for " + int(g_afk_time) + " seconds\n");
			} else if (timeLeft < 10) {
				g_PlayerFuncs.SayText(plr, "AFK - you will become a ball if you don't move within " + timeLeft + " seconds\n");
			}
		}
		if (plr.pev.targetname == "player_ball" || plr.pev.targetname == "pong_ball") {
			if (isAfk && !plr.IsAlive()) {
				g_Scheduler.SetTimeout("delay_respawn", 0.5f, EHandle(plr));
			}
		}
	}
	
	string statusMsg;
	for (int k = 0; k < 18; k++) {
		statusMsg += "Hole " + (k+1) + ":";
	
		string s1 = ball_count[k] > 1 ? "s" : "";
		string s2 = golfer_count[k] > 1 ? "s" : "";
	
		if (golfer_count[k] > 0 && ball_count[k] > 0)
			
			statusMsg += " " + ball_count[k] + " ball" + s1 + ", " + golfer_count[k] + " golfer" + s2 + "\n";
		else if (golfer_count[k] > 0) {
			statusMsg += " " + golfer_count[k] + " golfer" + s2 + "\n";
		} else if (ball_count[k] > 0) {
			statusMsg += " " + ball_count[k] + " ball" + s1 + "\n";
		} else {
			statusMsg += "\n";
		}
	}
	
	for ( int i = 1; i <= g_Engine.maxClients; i++ )
	{
		CBasePlayer@ plr = g_PlayerFuncs.FindPlayerByIndex(i);
		if (plr is null or !plr.IsConnected())
			continue;
		
		PlayerState@ state = getPlayerState(plr);
		if (state.showStats) {
			HUDTextParams params;
			params.y = 0.1;
			params.effect = 0;
			params.r1 = 255;
			params.g1 = 255;
			params.b1 = 255;
			params.fadeinTime = 0;
			params.fadeoutTime = 0;
			params.holdTime = 2.0f;
			params.x = 0.0;
			params.channel = 0;

			g_PlayerFuncs.HudMessage(plr, params, statusMsg);
		}
	}
}

HookReturnCode PlayerTakeDamage(DamageInfo@ info)
{
	CBasePlayer@ victim = cast<CBasePlayer@>(g_EntityFuncs.Instance(info.pVictim.pev));
	CBasePlayer@ attacker = cast<CBasePlayer@>(info.pAttacker);
	
	if (attacker !is null && attacker.IsPlayer()) {
		if (victim.pev.targetname == "player_ball" && attacker.pev.targetname == "golfer") {
			PlayerState@ state = getPlayerState(victim);
			state.addSwinger(attacker);
			
			if (state.h_arrow) {
				CBaseEntity@ arrow = state.h_arrow;
				arrow.pev.colormap = attacker.pev.colormap;
			}
			
			g_EngineFuncs.MakeVectors(attacker.pev.v_angle);
			victim.pev.velocity = g_Engine.v_forward*info.flDamage*20;
			
			state.swings += 1;
			
			float minDmg = 22;
			float maxDmg = 107;
			int power = (int(info.flDamage / maxDmg*100) / 10) * 10; // round down to nearest 10%
			
			HUDTextParams params;
			params.y = 1.0;
			params.x = -1;
			params.effect = 0;
			params.r1 = 255;
			params.g1 = 0;
			params.b1 = 0;
			params.fadeinTime = 0;
			params.fadeoutTime = 0.5f;
			params.holdTime = 1.0f;
			params.channel = 1;

			g_PlayerFuncs.HudMessage(attacker, params, "Power: " + power + "%");
			
		} else if (victim.pev.targetname == "golfer" && attacker.pev.targetname == "player_ball") {
			g_EngineFuncs.MakeVectors(attacker.pev.v_angle);
			
			g_EngineFuncs.MakeVectors(attacker.pev.v_angle);
			victim.pev.velocity = g_Engine.v_forward*info.flDamage*60;
		}
	}
	
	return HOOK_CONTINUE;
}

void hole_sound(CBaseEntity@ pActivator, CBaseEntity@ pCaller, USE_TYPE useType, float flValue) {
	dictionary keys;
	keys["origin"] = pActivator.pev.origin.ToString();
	keys["message"] = "minigolf/golfball.wav";
	keys["health"] = "10";
	keys["playmode"] = "5";
	keys["linearmin"] = "1";
	keys["linearmax"] = "8";
	keys["spawnflags"] = "16"; // start silent
	
	CBaseEntity@ ent = g_EntityFuncs.CreateEntity("ambient_generic", keys, true);
	ent.Use(null, null, USE_ON);
	g_EntityFuncs.Remove(ent);
}

void golfer_spawn(CBaseEntity@ pActivator, CBaseEntity@ pCaller, USE_TYPE useType, float flValue)
{
	if (pActivator !is null && pActivator.IsPlayer()) {
		CBasePlayer@ plr = cast<CBasePlayer@>(pActivator);
		plr.RemoveAllItems(false);
		plr.SetItemPickupTimes(0);
		plr.GiveNamedItem("weapon_pipewrench");
		plr.GiveNamedItem("weapon_medkit");
		PlayerState@ state = getPlayerState(plr);
		state.lastHole = "t_long18_ball";
	}
}

void reset_ball(CBaseEntity@ pActivator, CBaseEntity@ pCaller, USE_TYPE useType, float flValue)
{
	if (pActivator !is null && pActivator.IsPlayer()) {
		CBasePlayer@ plr = cast<CBasePlayer@>(pActivator);
		PlayerState@ state = getPlayerState(plr);
		state.reset(plr);
	}
}

void ball_tele(CBaseEntity@ pActivator, CBaseEntity@ pCaller, USE_TYPE useType, float flValue)
{
	CBasePlayer@ plr = cast<CBasePlayer@>(pActivator);
	if (plr is null)
		return;
		
	string thisHole = pCaller.pev.targetname;
	if (thisHole == "t_long11_ball2" || thisHole == "t_long11_ball_mm") {
		thisHole = "t_long11_ball";
	}
		
	PlayerState@ state = getPlayerState(plr);
	
	if (pActivator.pev.targetname == "golfer") {
		// too lazy to set up separate teleports for each class
		g_EngineFuncs.MakeVectors(pCaller.pev.angles);
		pActivator.pev.origin = pActivator.pev.origin - g_Engine.v_forward*64;
		pActivator.pev.origin.z += 16; // for hole 18
		state.lastHole = thisHole;
	} else if (pActivator.pev.targetname == "player_ball") {
		if (state.lastHole != thisHole) {
			// made it to the next hole
			state.finishHole(plr, thisHole);
		} else {
			// out of bounds!
		}
	}
}

int last_hit_count = 0;
bool game_over = false;
void pong_win_check(CBaseEntity@ pActivator, CBaseEntity@ pCaller, USE_TYPE useType, float flValue)
{
	int hit_count = 0;
	CBaseEntity@ ent = null;
	do {
		@ent = g_EntityFuncs.FindEntityByTargetname(ent, "pong1*");
		
		if (ent !is null)
		{
			if (ent.pev.effects == 128) {
				hit_count++;
			}
		}
	} while (ent !is null);
	
	if (hit_count >= 15) {
		g_EntityFuncs.FireTargets("win_pong_relay", null, null, USE_TOGGLE);
		
		CBasePlayer@ ballPlr = cast<CBasePlayer@>(pActivator);
		PlayerState@ state = getPlayerState(ballPlr);
		
		CBaseEntity@ pongController = @g_EntityFuncs.FindEntityByTargetname(null, "pong_golfer");
		if (pongController !is null && !game_over) {
			game_over = true;
			
			ballPlr.pev.frags += 20;
			pongController.pev.frags += 20;
			
			g_PlayerFuncs.SayTextAll(ballPlr, "" + pongController.pev.netname + " and " + ballPlr.pev.netname + " finished the last hole!\n");
			
			g_SoundSystem.PlaySound(ballPlr.edict(), CHAN_ITEM, "minigolf/fanfare.wav", 1.0f, 1.0f, 0, 100);
			g_SoundSystem.PlaySound(pongController.edict(), CHAN_ITEM, "minigolf/fanfare.wav", 1.0f, 1.0f, 0, 100);
		}
		
	}
	if (hit_count != last_hit_count) {
		last_hit_count = hit_count;
		g_EntityFuncs.FireTargets("pong_hit_woop", null, null, USE_TOGGLE);
	}
}

float last_unstuck_check = 0;
int stuck_time = 0;

HookReturnCode PlayerPreThink(CBasePlayer@ plr, uint& out) {

	if (plr.pev.targetname == "player_ball") {
		plr.pev.friction = 0.1;
		plr.pev.maxspeed = 20;
		plr.pev.gravity = 2;
	} else if (plr.pev.targetname == "pong_ball") {
		plr.pev.friction = 0.1;
		plr.pev.maxspeed = 0;
		plr.pev.gravity = 0.001;
		
		if (plr.pev.origin.z < 1636) { // hack to check if in the starting area or not
			if (g_Engine.time - last_unstuck_check > 1) {
				last_unstuck_check = g_Engine.time;
				
				if ((abs(plr.pev.velocity.y) < 50 || abs(plr.pev.velocity.z) < 50)) {
					if (stuck_time++ > 2) {
						stuck_time = 0;
						plr.pev.velocity.y = 250;
						plr.pev.velocity.z = 250;
					}
				} else {
					stuck_time = 0;
				}
			}
		}
	} else {
		plr.pev.maxspeed = 0;
		plr.pev.gravity = 1;
	}

	return HOOK_CONTINUE;
}

void delay_respawn(EHandle h_plr) {
	if (!h_plr.IsValid()) {
		return;
	}
	
	CBasePlayer@ plr = cast<CBasePlayer@>(h_plr.GetEntity());
	
	g_PlayerFuncs.RespawnPlayer(plr, true, true);
}

void reset_pong_blockers(CBasePlayer@ plr) {
	if (plr.pev.targetname == "pong_ball") {
		g_EntityFuncs.FireTargets("reset_pong_ball_block", null, null, USE_TOGGLE);
	}
	if (plr.pev.targetname == "pong_golfer") {
		g_EntityFuncs.FireTargets("reset_pong_golfer_block", null, null, USE_TOGGLE);
	}
}

bool doCommand(CBasePlayer@ plr, const CCommand@ args)
{
	PlayerState@ state = getPlayerState(plr);
	bool isAdmin = g_PlayerFuncs.AdminLevel(plr) >= ADMIN_YES;
	
	if ( args.ArgC() > 0 )
	{
		if (args[0] == ".ball" || args[0] == ".golfball")
		{
			if (plr.pev.targetname == "player_ball" || plr.pev.targetname == "pong_ball") {
				g_PlayerFuncs.SayText(plr, "You are already a golf ball.\n");
				return true;
			}
			
			reset_pong_blockers(plr);
			plr.pev.targetname = "player_ball";
			g_Scheduler.SetTimeout("delay_respawn", 0.1f, EHandle(plr));
			
			state.reset(plr);
			g_PlayerFuncs.SayText(plr, "You are a golf ball. Sit back and relax!\n");
			return true;
		}
		if (args[0] == ".golf" || args[0] == ".golfer")
		{
			if (plr.pev.targetname == "golfer" || plr.pev.targetname == "pong_golfer") {
				g_PlayerFuncs.SayText(plr, "You are already a golfer.\n");
				return true;
			}
			reset_pong_blockers(plr);
			plr.pev.targetname = "golfer";
			g_EntityFuncs.Remove(state.h_arrow);
			state.lastNotAfk = g_Engine.time;
			g_Scheduler.SetTimeout("delay_respawn", 0.1f, EHandle(plr));
			
			g_PlayerFuncs.SayText(plr, "You are now a golfer. Say '.status' to show ball locations.\n");
			
			return true;
		}
		if (args[0] == ".stat" || args[0] == ".stats" || args[0] == ".status" || args[0] == ".statsu")
		{
			state.showStats = !state.showStats;
			if (state.showStats) {
				g_PlayerFuncs.SayText(plr, "Showing ball/golfer status\n");
			} else {
				g_PlayerFuncs.SayText(plr, "Hiding ball/golfer status\n");
			}
			return true;
		}
	}
	
	return false;
}

HookReturnCode ClientSay( SayParameters@ pParams )
{
	CBasePlayer@ plr = pParams.GetPlayer();
	const CCommand@ args = pParams.GetArguments();	
	if (doCommand(plr, args))
	{
		pParams.ShouldHide = true;
		return HOOK_HANDLED;
	}
	return HOOK_CONTINUE;
}
